/*  FILE:			TWINE.CPP
	AUTHOR:			Brian A. Lowe, Jr.
	DATE:			March 12, 1995

	OVERVIEW
	========
	Twine object--an efficient and automatic string container class.
	This is the definition file for the Twine Object, a C++ class for storing, managing and deleting strings.

	This code is given freely to the public domain with no warranties whatsoever. Use it, improve it.
	Contact me on CompuServe at 76640,2664 or Internet at brianlow@mcs.com (preferred) with any
	praise or criticism, bugs or code enhancements you've made. It is some payback for code I've
	downloaded and used.

	Code is set with 4 spaces per tab.

	See TWINE.TXT for usage information and code examples.

	Definition of the Twine Object, a C++ class for storing, managing and deleting strings. A Twine is
	a single ended linked list object which stores strings. The beauty of the class is that it can be
	created as an automatic object, yet its strings are stored dynamically on the heap. When execution
	leaves the scope of the twine, the heap string memory is deleted as the Twine is automatically
	destroyed. The maximum number of strings is limited solely by the amount of available memory; you
	will have trouble maxing out this object; there is no 64K boundary on string count. The maximum string
	count is 2^(bits-in-long)/2 or over 2 billion with 32 bit longs; a negative index is not allowed. Each
	string can be up to 64K long and must be standard C++ null terminated.

	Strings can be added as exclusive or non-exclusive, so if you want to prevent having two copies of
	identical strings in the Twine, set bExclusive = TRUE. Strings can be added at the end of the list or
	by index. They can be removed by index or by string content. They can be replaced by index position or
	by string content. You can see if a Twine contains a particular string and can access the contents of
	a string by index as an array by the array bracket operators--Twine[iIndex]. Indexing is zero-based.
	You can clear the contents of a Twine.

	See TWINE.H for information on each class method and its return value and TWINE.TXT for information
	on using the class and its methods.
*/

#include <string.h>
#include "twine.h"

Twine::Twine()
{
	pFirstEntry = new StringHolder;
	pFirstEntry->pNext = NULL;
	pFirstEntry->lpszString = NULL;
	pFirstEntry->lIndex = 0L;
	pLastEntry = pFirstEntry;
}

Twine::~Twine()
{
	Clear();
	delete pLastEntry;	// Same as pFirstEntry after Clear() [also the *only* entry]
}

// The string to be added goes into the "current" list entry which was already created and initialized.
// A new list entry is then created which has no string allocated or assigned. This means the index for the
// pLastEntry list (pLastEntry.lIndex) is equal to the number of strings which have been added (because
// it is zero-based).
long Twine::Add (char far * lpszStringToAdd, BOOL bExclusive)
{	// Make sure constructor could allocate memory for StringHolder and string to be added is valid
	if ((pLastEntry == NULL) || (lpszStringToAdd == NULL))
		return -1L;
	if (bExclusive)
	{
		long lAlreadyThere = Contains (lpszStringToAdd);
		if (lAlreadyThere != -1L)
			return lAlreadyThere;										// Return index of string found in list
	}
	pLastEntry->lpszString = new char [strlen (lpszStringToAdd) + 1];	// Allocate memory for string
	if (pLastEntry->lpszString == NULL)
		return -1L;
	strcpy (pLastEntry->lpszString, lpszStringToAdd);					// Copy string to current list

	StringHolder * pNextEntry = new StringHolder;						// Make a new list
	if (pNextEntry == NULL)
		return -1L;
	pLastEntry->pNext = pNextEntry;										// Set current list "next" ptr to new list

	pNextEntry->pNext = NULL;								// Initialize entries in new list
	pNextEntry->lpszString = NULL;
	pNextEntry->lIndex = pLastEntry->lIndex + 1;
	pLastEntry = pNextEntry;											// Set last ptr to new list
	return pLastEntry->lIndex;
}

long Twine::Add (long lAddAtIndex, char far * lpszStringToAdd, BOOL bExclusive)
{	// Make sure constructor could allocate memory for StringHolder and lAddAtIndex & string to add are valid
	if ((pLastEntry == NULL) || (lAddAtIndex < 0L) || (lpszStringToAdd == NULL))
		return -1L;
	if (lAddAtIndex >= pLastEntry->lIndex)					// Index >= last index, so add string to end of list
		return Add (lpszStringToAdd, bExclusive);			// Explicitly pass in bExclusive so proper action is taken
	if (bExclusive)
	{
		long lAlreadyThere = Contains (lpszStringToAdd);
		if (lAlreadyThere != -1L)
			return lAlreadyThere;					// Return index of string found in list
	}
	StringHolder * pStringAtIndex = pFirstEntry, * pPrevString = NULL;
	BOOL bKeepLooking = TRUE;
	while ((pStringAtIndex != pLastEntry) && bKeepLooking)					// This never tests pLastEntry, which
	{																		// never holds a valid string
		if (pStringAtIndex->lIndex == lAddAtIndex)							// See if indices match
		{
			StringHolder * pThisEntry = new StringHolder;
			pThisEntry->lIndex = lAddAtIndex;
			pThisEntry->lpszString = new char [strlen (lpszStringToAdd) + 1];	// Allocate memory for string
			if (pThisEntry->lpszString == NULL)
		    	return -1L;
			strcpy (pThisEntry->lpszString, lpszStringToAdd);		// Copy string to current list

			if (pPrevString == NULL)					// If adding into location 0, update pFirstEntry
			{														// rather than previous entry's pNext ptr
				pThisEntry->pNext = pFirstEntry;
				pFirstEntry = pThisEntry;
			}
			else
			{
				pThisEntry->pNext = pPrevString->pNext;				// Point this one's next ptr to next List
				pPrevString->pNext = pThisEntry;					// Point prev one's next ptr to this one
			}
			while (pStringAtIndex != pLastEntry)					// Increment remaining indices except last
			{
				pStringAtIndex->lIndex++;
				pStringAtIndex = pStringAtIndex->pNext;
			}
			pStringAtIndex->lIndex++;								// Make sure pLastEntry.lIndex gets incremented
			bKeepLooking = FALSE;									// Terminate outer "while" loop
		}
		else		// Go to next list entry
        {
        	pPrevString = pStringAtIndex;
			pStringAtIndex = pStringAtIndex->pNext;
		}
	}
	return pLastEntry->lIndex;
}

BOOL Twine::Remove (long lRemoveAtIndex)
{
	if ((lRemoveAtIndex >= pLastEntry->lIndex) || (lRemoveAtIndex < 0L))
		return FALSE;
	StringHolder * pStringAtIndex = pFirstEntry, * pPrevString = pFirstEntry;
	BOOL bKeepLooking = TRUE;
	while ((pStringAtIndex != pLastEntry) && bKeepLooking)		// This never tests pLastEntry, which
	{															// does not yet hold a valid string
		if (pStringAtIndex->lIndex == lRemoveAtIndex)			// See if indices match
		{
			if (lRemoveAtIndex == 0L)							// If removing from location 0, update pFirstEntry
				pFirstEntry = pFirstEntry->pNext;
			else
				pPrevString->pNext = pStringAtIndex->pNext;	// Point prev one's next ptr to next one (skip this one)
			delete pStringAtIndex->lpszString;
			delete  pStringAtIndex;

			while (pPrevString != pLastEntry)
			{
				pPrevString = pPrevString->pNext;						// Point to next string list
				pPrevString->lIndex--;									// Decrement index
			}
			bKeepLooking = FALSE;										// Terminate outer "while" loop
		}
		else															// Go to next list entry
		{
			pPrevString = pStringAtIndex;
			pStringAtIndex = pStringAtIndex->pNext;
		}
	}
	return !bKeepLooking;
}

BOOL Twine::Remove (char far * lpszStringToRemove)
{
	if (lpszStringToRemove == NULL)						// Make sure string is valid
		return FALSE;
	StringHolder * pStringAtIndex = pFirstEntry, * pPrevString = pFirstEntry;
	BOOL bKeepLooking = TRUE;
	while ((pStringAtIndex != pLastEntry) && bKeepLooking)				// This never tests pLastEntry, which
	{																	// never holds a valid string
		if (!stricmp (pStringAtIndex->lpszString, lpszStringToRemove))	// See if strings match
		{
			if (pPrevString == NULL)						// If removing from location 0, update pFirstEntry
				pFirstEntry = pStringAtIndex->pNext;
			else
				pPrevString->pNext = pStringAtIndex->pNext;	// Point prev one's next ptr to next one (skip this one)
			delete pStringAtIndex->lpszString;
			delete  pStringAtIndex;

			while (pPrevString != pLastEntry)
			{
				pPrevString = pPrevString->pNext;					// Point to next string list
				pPrevString->lIndex--;								// Decrement index
			}
			bKeepLooking = FALSE;									// Terminate outer "while" loop
		}
		else														// Go to next list entry
		{
			pPrevString = pStringAtIndex;
			pStringAtIndex = pStringAtIndex->pNext;
		}
	}
	return !bKeepLooking;
}

BOOL Twine::Replace (char far * lpszNewString, long lReplaceAtIndex)
{
	if ((lReplaceAtIndex >= pLastEntry->lIndex) || (lReplaceAtIndex < 0L) || (lpszNewString == NULL))
		return FALSE;
	StringHolder * pStringAtIndex = pFirstEntry;
	BOOL bKeepLooking = TRUE;
	while ((pStringAtIndex != pLastEntry) && bKeepLooking)			// This never tests pLastEntry, which
	{																// never holds a valid string
		if (pStringAtIndex->lIndex == lReplaceAtIndex)				// See if indices match
		{
			char far * lpszTemp = new char [strlen (lpszNewString) + 1];	// To be able to keep current data when
			if (lpszTemp == NULL)									// memory can't be allocated, try to allocate
				return FALSE;										// before deleting current string
			else
			{
				strcpy (lpszTemp, lpszNewString);					// Memory OK so copy string to new location
				delete pStringAtIndex->lpszString;					// Delete existing string
				pStringAtIndex->lpszString = lpszTemp;				// Point to new string
				bKeepLooking = FALSE;
			}
		}
		else											// Go to next list entry
			pStringAtIndex = pStringAtIndex->pNext;
	}
	return !bKeepLooking;
}

BOOL Twine::Replace (char far * lpszStringToReplace, char far * lpszNewString)
{
	if ((lpszStringToReplace == NULL) || (lpszNewString == NULL))
		return FALSE;
	StringHolder * pStringAtIndex = pFirstEntry;
	BOOL bKeepLooking = TRUE;
	while ((pStringAtIndex != pLastEntry) && bKeepLooking)				// This never tests pLastEntry, which
	{																	// never holds a valid string
		if (!stricmp (pStringAtIndex->lpszString, lpszStringToReplace))	// See if matching string found
			pStringAtIndex = pStringAtIndex->pNext;						// Go to next list entry
		else
		{																// Replace string
			delete pStringAtIndex->lpszString;
			pStringAtIndex->lpszString = new char [strlen (lpszNewString) + 1];
			strcpy (pStringAtIndex->lpszString, lpszNewString);
			bKeepLooking = FALSE;
		}
	}
	return !bKeepLooking;
}

long Twine::Contains (char far * lpszStringToFind)
{
	if (lpszStringToFind == NULL)
		return FALSE;
	StringHolder * pStringAtIndex = pFirstEntry;

	while (pStringAtIndex != pLastEntry)								// This never tests pLastEntry, which
	{																	// never holds a valid string
		if (!stricmp (pStringAtIndex->lpszString, lpszStringToFind))	// See if matching string found
			return pStringAtIndex->lIndex;
		else															// Go to next list entry
			pStringAtIndex = pStringAtIndex->pNext;
	}
	return -1L;
}

void Twine::Clear()
{
	if (pFirstEntry == pLastEntry)										// Already cleared
		return;
	StringHolder * pStringAtIndex = pFirstEntry, * pNextEntry;

	while (pStringAtIndex != pLastEntry)								// This never deletes pLastEntry, which
	{																	// becomes the only entry after Clear()
		pNextEntry = pStringAtIndex->pNext;
		delete pStringAtIndex->lpszString;
		delete  pStringAtIndex;
		pStringAtIndex = pNextEntry;
	}
	pFirstEntry = pLastEntry;											// Reset pFirstEntry
	pLastEntry->lIndex = 0L;
}

char far * Twine::operator[] (long lIndexToFind)
{
	if ((lIndexToFind >= pLastEntry->lIndex) || (lIndexToFind < 0L))
		return NULL;
	StringHolder * pStringAtIndex = pFirstEntry;

	while (pStringAtIndex != pLastEntry)								// This never tests pLastEntry, which
	{																	// never holds a valid string
		if (pStringAtIndex->lIndex == lIndexToFind)						// See if indices match
			return pStringAtIndex->lpszString;
		else															// Go to next list entry if no match
			pStringAtIndex = pStringAtIndex->pNext;
	}
	return NULL;
}

